@@ -1,25 +1,10 @@
 """
     Python WebDAV Server.
-    Copyright (C) 1999 Christian Scholz (ruebe@aachen.heimat.de)
 
-    This library is free software; you can redistribute it and/or
-    modify it under the terms of the GNU Library General Public
-    License as published by the Free Software Foundation; either
-    version 2 of the License, or (at your option) any later version.
+    This module builds on AuthServer by implementing the standard DAV
+    methods.
 
-    This library is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-    Library General Public License for more details.
-
-    You should have received a copy of the GNU Library General Public
-    License along with this library; if not, write to the Free
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
-
-This module builds on AuthServer by implementing the standard DAV
-methods.
-
-Subclass this class and specify an IFACE_CLASS. See example.
+    Subclass this class and specify an IFACE_CLASS. See example.
 
 """
 
@@ -41,6 +26,7 @@ import AuthServer
 import urlparse
 import urllib
 import random
+import logging
 
 from propfind import PROPFIND
 from report import REPORT
@@ -48,7 +34,7 @@ from delete import DELETE
 from davcopy import COPY
 from davmove import MOVE
 
-from utils import rfc1123_date
+from utils import rfc1123_date, IfParser, tokenFinder
 from string import atoi,split
 from errors import *
 
@@ -57,9 +43,13 @@ from locks import LockManager
 import gzip
 import StringIO
 
+log = logging.getLogger(__name__)
+
+BUFFER_SIZE = 128 * 1000 # 128 Ko
+
 class DAVRequestHandler(AuthServer.BufferedAuthRequestHandler, LockManager):
-    """Simple DAV request handler with 
-    
+    """Simple DAV request handler with
+
     - GET
     - HEAD
     - PUT
@@ -75,33 +65,35 @@ class DAVRequestHandler(AuthServer.BufferedAuthRequestHandler, LockManager):
 
     It uses the resource/collection classes for serving and
     storing content.
-    
+
     """
 
     server_version = "DAV/" + __version__
     encode_threshold = 1400 # common MTU
 
-    ### utility functions
-    def _log(self, message):
-        pass
-
-    def send_body(self,DATA,code,msg,desc,ctype='application/octet-stream',headers={}):
+    def send_body(self, DATA, code = None, msg = None, desc = None, ctype='application/octet-stream', headers={}):
         """ send a body in one part """
+        log.debug("Use send_body method")
 
         self.send_response(code,message=msg)
         self.send_header("Connection", "close")
         self.send_header("Accept-Ranges", "bytes")
         self.send_header('Date', rfc1123_date())
-        
+        self.send_header('DAV', DAV_VERSION_2['version'])
+
         for a,v in headers.items():
             self.send_header(a,v)
-        
+
         if DATA:
             if 'gzip' in self.headers.get('Accept-Encoding', '').split(',') \
                     and len(DATA) > self.encode_threshold:
                 buffer = StringIO.StringIO()
                 output = gzip.GzipFile(mode='wb', fileobj=buffer)
-                output.write(DATA)
+                if isinstance(DATA, str) or isinstance(DATA, unicode):
+                    output.write(DATA)
+                else:
+                    for buf in DATA:
+                        output.write(buf)
                 output.close()
                 buffer.seek(0)
                 DATA = buffer.getvalue()
@@ -111,38 +103,105 @@ class DAVRequestHandler(AuthServer.BufferedAuthRequestHandler, LockManager):
             self.send_header('Content-Type', ctype)
         else:
             self.send_header('Content-Length', 0)
-        
+
         self.end_headers()
         if DATA:
-            self._append(DATA)
+            if isinstance(DATA, str) or isinstance(DATA, unicode):
+                log.debug("Don't use iterator")
+                self._append(DATA)
+            else:
+                if self._config.DAV.getboolean('http_response_use_iterator'):
+                    # Use iterator to reduce using memory
+                    log.debug("Use iterator")
+                    self.wfile.write(self._buffer)
+                    self.wfile.flush()
+                    self._buffer=""
+                    for buf in DATA:
+                        self.wfile.write(buf)
+                        self.wfile.flush()
+                else:
+                    # Don't use iterator, it's a compatibility option
+                    log.debug("Don't use iterator")
+                    self._append(DATA.read())
 
-    def send_body_chunks(self,DATA,code,msg,desc,ctype='text/xml; encoding="utf-8"'):
+    def send_body_chunks_if_http11(self, DATA, code, msg = None, desc = None, 
+                                   ctype='text/xml; encoding="utf-8"', 
+                                   headers={}):
+        if (
+            self.request_version == 'HTTP/1.0' or
+            not self._config.DAV.getboolean('chunked_http_response')
+        ):
+            self.send_body(DATA, code, msg, desc, ctype, headers)
+        else:
+            self.send_body_chunks(DATA, code, msg, desc, ctype, headers)
+ 
+    def send_body_chunks(self, DATA, code, msg=None, desc=None, 
+                            ctype='text/xml"', headers={}):
         """ send a body in chunks """
-        
+
         self.responses[207]=(msg,desc)
         self.send_response(code,message=msg)
         self.send_header("Content-type", ctype)
-        self.send_header("Connection", "close")
         self.send_header("Transfer-Encoding", "chunked")
         self.send_header('Date', rfc1123_date())
+        self.send_header('DAV', DAV_VERSION_2['version'])
 
-        if 'gzip' in self.headers.get('Accept-Encoding', '').split(',') \
-                and len(DATA) > self.encode_threshold:
+        for a,v in headers.items():
+            self.send_header(a,v)
+
+        if DATA:
+            if (
+                ('gzip' in self.headers.get('Accept-Encoding', '').split(',')) and
+                (len(DATA) > self.encode_threshold)
+            ):
                 buffer = StringIO.StringIO()
                 output = gzip.GzipFile(mode='wb', fileobj=buffer)
-                output.write(DATA)
+                if isinstance(DATA, str):
+                    output.write(DATA)
+                else:
+                    for buf in DATA:
+                        output.write(buf)
                 output.close()
                 buffer.seek(0)
                 DATA = buffer.getvalue()
                 self.send_header('Content-Encoding', 'gzip')
 
+            self.send_header('Content-Length', len(DATA))
+            self.send_header('Content-Type', ctype)
+
+        else:
+            self.send_header('Content-Length', 0)
+
         self.end_headers()
 
-        self._append(hex(len(DATA))[2:]+"\r\n")
-        self._append(DATA)
-        self._append("\r\n")
-        self._append("0\r\n")
-        self._append("\r\n")
+        if DATA:
+            if isinstance(DATA, str) or isinstance(DATA, unicode):
+                self._append(hex(len(DATA))[2:]+"\r\n")
+                self._append(DATA)
+                self._append("\r\n")
+                self._append("0\r\n")
+                self._append("\r\n")
+            else:
+                if self._config.DAV.getboolean('http_response_use_iterator'):
+                    # Use iterator to reduce using memory
+                    self.wfile.write(self._buffer)
+                    self.wfile.flush()
+                    self._buffer=""
+
+                    for buf in DATA:
+                        self.wfile.write(hex(len(buf))[2:]+"\r\n")
+                        self.wfile.write(buf)
+                        self.wfile.write("\r\n")
+
+                    self.wfile.write("0\r\n")
+                    self.wfile.write("\r\n")
+                else:
+                    # Don't use iterator, it's a compatibility option
+                    self._append(hex(len(DATA))[2:]+"\r\n")
+                    self._append(DATA.read())
+                    self._append("\r\n")
+                    self._append("0\r\n")
+                    self._append("\r\n")
 
     ### HTTP METHODS called by the server
 
@@ -152,9 +211,9 @@ class DAVRequestHandler(AuthServer.BufferedAuthRequestHandler, LockManager):
         self.send_response(200)
         self.send_header("Content-Length", 0)
 
-        if self._config.DAV.lockemulation is True:
-            if self._config.DAV.verbose is True:
-                print >>sys.stderr, 'Activated LOCK,UNLOCK emulation (experimental)'
+        if self._config.DAV.getboolean('lockemulation') is True:
+            if self._config.DAV.getboolean('verbose') is True:
+                log.info('Activated LOCK,UNLOCK emulation (experimental)')
 
             self.send_header('Allow', DAV_VERSION_2['options'])
             self.send_header('DAV', DAV_VERSION_2['version'])
@@ -190,18 +249,35 @@ class DAVRequestHandler(AuthServer.BufferedAuthRequestHandler, LockManager):
             content_type = dc.get_prop(uri,"DAV:","getcontenttype")
         except DAV_NotFound: content_type = "application/octet-stream"
 
+        range = None
+        status_code = 200
+        if 'Range' in self.headers:
+            p = self.headers['Range'].find("bytes=")
+            if p != -1:
+                range = self.headers['Range'][p + 6:].split("-")
+                status_code = 206
+
         # get the data
         try:
-            data = dc.get_data(uri)
+            data = dc.get_data(uri, range)
         except DAV_Error, (ec,dd):
             self.send_status(ec)
-            return 
+            return ec
 
         # send the data
         if with_body is False:
             data = None
 
-        self.send_body(data, '200', "OK", "OK", content_type, headers)
+        if isinstance(data, str) or isinstance(data, unicode):
+            self.send_body(data, status_code, None, None, content_type, headers)
+        else:
+            headers['Keep-Alive'] = 'timeout=15, max=86'
+            headers['Connection'] = 'Keep-Alive'
+            self.send_body_chunks_if_http11(data, status_code, None, None, 
+                                            content_type, headers)
+
+        return status_code
+            
 
     def do_HEAD(self):
         """ Send a HEAD response: Retrieves resource information w/o body """
@@ -211,7 +287,17 @@ class DAVRequestHandler(AuthServer.BufferedAuthRequestHandler, LockManager):
     def do_GET(self):
         """Serve a GET request."""
 
-        return self._HEAD_GET(with_body=True)
+        log.debug(self.headers)
+
+        try:
+            status_code = self._HEAD_GET(with_body=True)
+            self.log_request(status_code)
+            return status_code
+        except IOError, e:
+            if e.errno == 32:
+                self.log_request(206)
+            else:
+                raise
 
     def do_TRACE(self):
         """ This will always fail because we can not reproduce HTTP requests. 
@@ -255,7 +341,7 @@ class DAVRequestHandler(AuthServer.BufferedAuthRequestHandler, LockManager):
             DATA = DATA.replace('<ns0:creationdate xmlns:ns0="DAV:">',
                                     '<ns0:creationdate xmlns:n="DAV:" xmlns:b="urn:uuid:c2f41010-65b3-11d1-a29f-00aa00c14882/" b:dt="dateTime.tz">')
 
-        self.send_body_chunks(DATA, '207','Multi-Status','Multiple responses')
+        self.send_body_chunks_if_http11(DATA, '207','Multi-Status','Multiple responses')
 
     def do_REPORT(self):
         """ Query properties on defined resource. """
@@ -272,14 +358,14 @@ class DAVRequestHandler(AuthServer.BufferedAuthRequestHandler, LockManager):
         uri = urlparse.urljoin(self.get_baseuri(dc), self.path)
         uri = urllib.unquote(uri)
 
-        rp = REPORT(uri, dc, self.headers.get('Depth', 'infinity'), body)
+        rp = REPORT(uri, dc, self.headers.get('Depth', '0'), body)
 
         try:
             DATA = '%s\n' % rp.createResponse()
         except DAV_Error, (ec,dd):
             return self.send_status(ec)
 
-        self.send_body_chunks(DATA, '207','Multi-Status','Multiple responses')
+        self.send_body_chunks_if_http11(DATA, '207','Multi-Status','Multiple responses')
 
     def do_MKCOL(self):
         """ create a new collection """
@@ -291,7 +377,9 @@ class DAVRequestHandler(AuthServer.BufferedAuthRequestHandler, LockManager):
         try:
             dc.mkcol(uri)
             self.send_status(201)
+            self.log_request(201)
         except DAV_Error, (ec,dd):
+            self.log_request(ec)
             return self.send_status(ec)
 
     def do_DELETE(self):
@@ -320,6 +408,7 @@ class DAVRequestHandler(AuthServer.BufferedAuthRequestHandler, LockManager):
                         break
             if not test:
                 self.send_status(412)
+                self.log_request(412)
                 return
 
         # Handle If-None-Match
@@ -341,6 +430,7 @@ class DAVRequestHandler(AuthServer.BufferedAuthRequestHandler, LockManager):
                         break
             if not test:
                 self.send_status(412)
+                self.log_request(412)
                 return
 
         # locked resources are not allowed to delete
@@ -350,25 +440,34 @@ class DAVRequestHandler(AuthServer.BufferedAuthRequestHandler, LockManager):
         dl = DELETE(uri,dc)
         if dc.is_collection(uri):
             res=dl.delcol()
-        else: res=dl.delone()
+            if res:
+                self.send_status(207,body=res)
+            else:
+                self.send_status(204)
+        else:
+            res=dl.delone() or 204
+            self.send_status(res)
 
-        if res:
-            self.send_status(207,body=res)
-        else: self.send_status(204)
 
     def do_PUT(self):
         dc=self.IFACE_CLASS
         uri=urlparse.urljoin(self.get_baseuri(dc), self.path)
         uri=urllib.unquote(uri)
 
+        log.debug("do_PUT: uri = %s" % uri)
+        log.debug('do_PUT: headers = %s' % self.headers)
         # Handle If-Match
         if self.headers.has_key('If-Match'):
+            log.debug("do_PUT: If-Match %s" % self.headers['If-Match'])
             test = False
             etag = None
             try:
                 etag = dc.get_prop(uri, "DAV:", "getetag")
             except:
                 pass
+
+            log.debug("do_PUT: etag = %s" % etag)
+
             for match in self.headers['If-Match'].split(','):
                 if match == '*':
                     if dc.exists(uri):
@@ -380,16 +479,22 @@ class DAVRequestHandler(AuthServer.BufferedAuthRequestHandler, LockManager):
                         break
             if not test:
                 self.send_status(412)
+                self.log_request(412)
                 return
 
         # Handle If-None-Match
         if self.headers.has_key('If-None-Match'):
+            log.debug("do_PUT: If-None-Match %s" % self.headers['If-None-Match'])
+
             test = True
             etag = None
             try:
                 etag = dc.get_prop(uri, "DAV:", "getetag")
             except:
                 pass
+
+            log.debug("do_PUT: etag = %s" % etag)
+
             for match in self.headers['If-None-Match'].split(','):
                 if match == '*':
                     if dc.exists(uri):
@@ -401,8 +506,38 @@ class DAVRequestHandler(AuthServer.BufferedAuthRequestHandler, LockManager):
                         break
             if not test:
                 self.send_status(412)
+                self.log_request(412)
                 return
 
+        # locked resources are not allowed to be overwritten
+        ifheader = self.headers.get('If')
+        if (
+            (self._l_isLocked(uri)) and
+            (not ifheader)
+        ):
+            return self.send_body(None, '423', 'Locked', 'Locked')
+
+        if ((self._l_isLocked(uri)) and (ifheader)):
+            uri_token = self._l_getLockForUri(uri)
+            taglist = IfParser(ifheader)
+            found = False
+            for tag in taglist:
+                for listitem in tag.list:
+                    token = tokenFinder(listitem)
+                    if (
+                        token and
+                        (self._l_hasLock(token)) and
+                        (self._l_getLock(token) == uri_token)
+                    ):
+                        found = True
+                        break
+                if found:
+                    break
+            if not found:
+                res = self.send_body(None, '423', 'Locked', 'Locked')
+                self.log_request(423)
+                return res
+
         # Handle expect
         expect = self.headers.get('Expect', '')
         if (expect.lower() == '100-continue' and
@@ -411,35 +546,77 @@ class DAVRequestHandler(AuthServer.BufferedAuthRequestHandler, LockManager):
             self.send_status(100)
             self._flush()
 
-        # read the body
-        body=None
-        if self.headers.has_key("Content-Length"):
-            l=self.headers['Content-Length']
-            body=self.rfile.read(atoi(l))
-
-        # locked resources are not allowed to be overwritten
-        if self._l_isLocked(uri):
-            return self.send_body(None, '423', 'Locked', 'Locked')
-
-        ct=None
+        content_type = None
         if self.headers.has_key("Content-Type"):
-            ct=self.headers['Content-Type']
-        try:
-            location = dc.put(uri,body,ct)
-        except DAV_Error, (ec,dd):
-            return self.send_status(ec)
+            content_type = self.headers['Content-Type']
 
         headers = {}
-        if location:
-            headers['Location'] = location
+        headers['Location'] = uri
 
         try:
-            etag = dc.get_prop(location or uri, "DAV:", "getetag")
+            etag = dc.get_prop(uri, "DAV:", "getetag")
             headers['ETag'] = etag
         except:
             pass
 
-        self.send_body(None, '201', 'Created', '', headers=headers)
+        expect = self.headers.get('transfer-encoding', '')
+        if (
+            expect.lower() == 'chunked' and
+            self.protocol_version >= 'HTTP/1.1' and
+            self.request_version >= 'HTTP/1.1'
+        ):
+            self.send_body(None, '201', 'Created', '', headers=headers)
+            self._flush()
+
+            dc.put(uri, self._readChunkedData(), content_type)
+        else:
+            # read the body
+            body=None
+            if self.headers.has_key("Content-Length"):
+                l=self.headers['Content-Length']
+                log.debug("do_PUT: Content-Length = %s" % l)
+                body=self._readNoChunkedData(atoi(l))
+            else:
+                log.debug("do_PUT: Content-Length = empty")
+
+            try:
+                dc.put(uri, body, content_type)
+            except DAV_Error, (ec,dd):
+                return self.send_status(ec)
+
+            self.send_body(None, '201', 'Created', '', headers=headers)
+            self.log_request(201)
+
+    def _readChunkedData(self):
+        l = int(self.rfile.readline(), 16)
+        while l > 0:
+            buf = self.rfile.read(l)
+            yield buf
+            self.rfile.readline()
+            l = int(self.rfile.readline(), 16)
+
+    def _readNoChunkedData(self, content_length):
+        if self._config.DAV.getboolean('http_request_use_iterator'):
+            # Use iterator to reduce using memory
+            return self.__readNoChunkedDataWithIterator(content_length)
+        else:
+            # Don't use iterator, it's a compatibility option
+            return self.__readNoChunkedDataWithoutIterator(content_length)
+        
+    def __readNoChunkedDataWithIterator(self, content_length):
+        while True:
+            if content_length > BUFFER_SIZE:
+                buf = self.rfile.read(BUFFER_SIZE)
+                content_length -= BUFFER_SIZE
+                yield buf
+            else:
+                buf = self.rfile.read(content_length)
+                yield buf
+                break
+
+    def __readNoChunkedDataWithoutIterator(self, content_length):
+        return self.rfile.read(content_length)
+
 
     def do_COPY(self):
         """ copy one resource to another """
@@ -486,12 +663,12 @@ class DAVRequestHandler(AuthServer.BufferedAuthRequestHandler, LockManager):
         d="infinity"
         if self.headers.has_key("Depth"):
             d=self.headers['Depth']
-            
-            if d!="0" and d!="infinity": 
+
+            if d!="0" and d!="infinity":
                 self.send_status(400)
                 return
-            
-            if d=="0":  
+
+            if d=="0":
                 res=cp.single_action()
                 self.send_status(res)
                 return
@@ -511,7 +688,7 @@ class DAVRequestHandler(AuthServer.BufferedAuthRequestHandler, LockManager):
                 return
 
         if res:
-            self.send_body_chunks(res,207,self.responses[207][0],
+            self.send_body_chunks_if_http11(res,207,self.responses[207][0],
                             self.responses[207][1],
                             ctype='text/xml; charset="utf-8"')
         else:
@@ -525,8 +702,10 @@ class DAVRequestHandler(AuthServer.BufferedAuthRequestHandler, LockManager):
     def send_status(self,code=200,mediatype='text/xml;  charset="utf-8"', \
                                 msg=None,body=None):
 
-        if not msg: msg=self.responses[code][1]
-        self.send_body(body,code,self.responses[code][0],msg,mediatype)
+        if not msg:
+            msg=self.responses.get(code, ['', ''])[1]
+
+        self.send_body(body,code,self.responses.get(code, [''])[0],msg,mediatype)
 
     def get_baseuri(self, dc):
         baseuri = dc.baseuri
@@ -535,3 +714,8 @@ class DAVRequestHandler(AuthServer.BufferedAuthRequestHandler, LockManager):
             uparts[1] = self.headers['Host']
             baseuri = urlparse.urlunparse(uparts)
         return baseuri
+
+    def log_message(self, *args):
+        AuthServer.BufferedAuthRequestHandler.log_message(self, 
+            *tuple(('- %s - ' + args[0],) + (self.headers.get('User-Agent', '?'),) + args[1:])
+        )
